@@ -2,7 +2,6 @@ package altsrc
 
 import (
 	"flag"
-	"io/ioutil"
 	"os"
 	"testing"
 
@@ -11,7 +10,7 @@ import (
 
 const (
 	fileName   = "current.json"
-	simpleJSON = `{"test": 15}`
+	simpleJSON = `{"test": 15, "testb": false}`
 	nestedJSON = `{"top": {"test": 15}}`
 )
 
@@ -34,14 +33,19 @@ func TestCommandJSONFileTest(t *testing.T) {
 		Action: func(c *cli.Context) error {
 			val := c.Int("test")
 			expect(t, val, 15)
+
+			valb := c.Bool("testb")
+			expect(t, valb, false)
 			return nil
 		},
 		Flags: []cli.Flag{
 			NewIntFlag(&cli.IntFlag{Name: "test"}),
-			&cli.StringFlag{Name: "load"}},
+			&cli.StringFlag{Name: "load"},
+			NewBoolFlag(&cli.BoolFlag{Name: "testb", Value: true}),
+		},
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -76,7 +80,7 @@ func TestCommandJSONFileTestGlobalEnvVarWins(t *testing.T) {
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
 
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -111,7 +115,7 @@ func TestCommandJSONFileTestGlobalEnvVarWinsNested(t *testing.T) {
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
 
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -143,7 +147,7 @@ func TestCommandJSONFileTestSpecifiedFlagWins(t *testing.T) {
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
 
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -175,7 +179,7 @@ func TestCommandJSONFileTestSpecifiedFlagWinsNested(t *testing.T) {
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
 
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -207,7 +211,7 @@ func TestCommandJSONFileTestDefaultValueFileWins(t *testing.T) {
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
 
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -239,7 +243,7 @@ func TestCommandJSONFileTestDefaultValueFileWinsNested(t *testing.T) {
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
 
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -273,7 +277,7 @@ func TestCommandJSONFileFlagHasDefaultGlobalEnvJSONSetGlobalEnvWins(t *testing.T
 			&cli.StringFlag{Name: "load"}},
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
@@ -307,13 +311,13 @@ func TestCommandJSONFileFlagHasDefaultGlobalEnvJSONSetGlobalEnvWinsNested(t *tes
 			&cli.StringFlag{Name: "load"}},
 	}
 	command.Before = InitInputSourceWithContext(command.Flags, NewJSONSourceFromFlagFunc("load"))
-	err := command.Run(c)
+	err := command.Run(c, test...)
 
 	expect(t, err, nil)
 }
 
 func writeTempFile(t *testing.T, name string, content string) func() {
-	if err := ioutil.WriteFile(name, []byte(content), 0666); err != nil {
+	if err := os.WriteFile(name, []byte(content), 0666); err != nil {
 		t.Fatalf("cannot write %q: %v", name, err)
 	}
 	return func() {
